#include "mainwindow.h"
#include "./ui_mainwindow.h"


MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent)
    , ui(std::make_unique<Ui::MainWindow>())
    , m_client(QUrl(QStringLiteral("ws://localhost:1234"))) // default server URL
    , m_username()
    , m_indexes()
    , m_actions()
{
    ui->setupUi(this);
    this->setWindowTitle("Generic chat application.exe");

    connect(&m_client, &Client::message_received,
            this, &MainWindow::on_messageReceived);
    connect(&m_client, &Client::disconnected,
            this, &MainWindow::on_clientDisconnected);
    connect(&m_client, &Client::connected,
            this, &MainWindow::on_clientConnected);
    resetChats();
    m_actions = ui->menuFile->actions();
    ui->messages->append("Register or login using Menu");
    ui->messages->append("If you are not connected to server automatically, set server URL");
}

MainWindow::~MainWindow()
{
    // do proper logout to clear the stackedwidget, avoids memory issues with non-explicit mm
    on_actionLogout_triggered();
}

void MainWindow::on_clientDisconnected()
{
    displayInfo("No connection to server. Reconnecting...");
}

void MainWindow::on_clientConnected()
{
    displayInfo("Connected to server.");
}

void MainWindow::on_sendButton_clicked()
{
    QString input = ui->lineEdit->text();
    ui->lineEdit->clear();

    // end here if no contact selected
    if (ui->receiver->text().isEmpty())
        return;

    // showing message for the sender
    QLocale loc;
    static_cast<QTextEdit *>(ui->messagesWidget->currentWidget())->
            append("<span style=color:grey>" + loc.toString(QDateTime::currentDateTime(), QLocale::ShortFormat) +
                   " " + "<span style=font-weight:bold>" + m_username + ": " + input + "</span></span>");

    m_client.add_message(Message::create_text_message(m_username, ui->receiver->text(), input));
}

QListWidgetItem *MainWindow::findContact(QString name)
{
    auto temp = ui->contacts->findItems(name, Qt::MatchExactly);
    if (temp.empty()) {
        return nullptr;
    }
    return temp.first();
}

void MainWindow::on_messageReceived(const Message &message)
{
    switch(message.get_type()) {
    case Message::Type::TEXT:
    {
        // ignore unknown senders
        if (m_indexes.find(message.get_sender()) == m_indexes.end())
            return;

        QLocale loc;
        static_cast<QTextEdit *>(ui->messagesWidget->widget(m_indexes[message.get_sender()]))->
                append(loc.toString(message.get_time(), QLocale::ShortFormat) + " <span style=font-weight:bold>" +
                       message.get_sender() + ": " +
                       message.get("message").toString() + "</span>");

        if (m_indexes[message.get_sender()] != ui->messagesWidget->currentIndex()) {
            QListWidgetItem *sender = findContact(message.get_sender());
            if (!sender) {
                displayInfo("ERROR: Contact missing");
                return;
            }

            // notify user by making the contact name bold
            QFont notificationFont = sender->font();
            notificationFont.setBold(true);
            sender->setFont(notificationFont);
        }
        break;
    }
    case Message::Type::FRIEND_REQUEST_RESPONSE:
    {
        if (!message.get("accepted").toBool())
            displayInfo("Friend request rejected by server.");
        break;
    }
    case Message::Type::CONTACT_LIST_RESPONSE:
    {
        QJsonValue name;
        foreach (name, message.get_contact_list()) {
            if (!findContact(name.toString())) {
                ui->contacts->addItem(name.toString());

                chats.push_back(std::make_unique<QTextEdit>(this));
                chats.back()->setReadOnly(true);
                m_indexes[name.toString()] = ui->messagesWidget->addWidget(chats.back().get());
            }
        }
        break;
    }
    case Message::Type::INFO:
    {
        Message initialMessage = message.get_initial_message();
        if (message.get("success").toBool()) {
            if (initialMessage.get_type() == Message::Type::LOGIN) {
                m_username = initialMessage.get("sender").toString();
                changeLoginState(true);
                ui->nameLabel->setText("Logged in as: " + m_username);
            }
            else if (initialMessage.get_type() == Message::Type::REGISTER) {
                QString name = initialMessage.get("sender").toString();
                QString password = initialMessage.get("password").toString();
                m_client.add_message(Message::create_login_message(name, password));
            }
        }
        displayInfo("SERVER: " + message.get("info").toString());
        break;
    }
    default:
        break;
    }
}

void MainWindow::on_lineEdit_returnPressed()
{
    on_sendButton_clicked();
}

void MainWindow::on_actionLogin_triggered()
{
    Login loginDialog(this);
    connect(&loginDialog, &Login::login_clicked,
            this, &MainWindow::on_loginFinished);
    loginDialog.exec();
}

void MainWindow::on_loginFinished(const QString &name, const QString &password)
{
    m_client.add_message(Message::create_login_message(name, password));
}

void MainWindow::on_addContactButton_clicked()
{
    QString contactName = QInputDialog::getText(this, tr("New contact"),
                                                tr("Requesting contact name:"));
    if (contactName.isEmpty())
        return;

    m_client.add_message(Message::create_friend_request(m_username, contactName));
}

void MainWindow::on_removeContactButton_clicked()
{
    auto *selected = ui->contacts->currentItem();
    if (!selected)
        return;

    if (QMessageBox::Yes == QMessageBox::question(this, tr("Remove contact"),
                                                  tr("Requesting confirmation."))) {
        m_client.add_message(Message::create_unfriend_request(m_username, selected->text()));
        m_client.add_message(Message::create_text_message(m_username, selected->text(),
                                                          "You have been removed from this user's contact list, "
                                                          "your messages are blocked."));
        m_indexes.erase(selected->text());
        ui->contacts->takeItem(ui->contacts->currentRow());
    }
}

void MainWindow::on_actionRegister_triggered()
{
    Register registerDialog(this);
    connect(&registerDialog, &Register::register_clicked,
            this, &MainWindow::on_registerFinished);
    registerDialog.exec();
}

void MainWindow::on_registerFinished(const QString &name, const QString &password)
{
    m_client.add_message(Message::create_register_message(name, password));
    ui->messages->append("If registration fails, try a different username");
}

void MainWindow::on_actionLogout_triggered()
{
    m_username = "";
    ui->receiver->clear();
    ui->messages->clear();
    ui->contacts->clear();
    ui->nameLabel->clear();
    for (int i = 0; i < ui->messagesWidget->count(); ++i)
        if (ui->messagesWidget->widget(i) != ui->messages)
            ui->messagesWidget->removeWidget(ui->messagesWidget->widget(i));

    resetChats();
    m_client.add_message(Message::create_logout_message(m_username));
    changeLoginState(false);
}

void MainWindow::on_actionSet_server_URL_triggered()
{
    QString urlString = QInputDialog::getText(this, tr("New URL"),
                                              tr("Enter new URL and port:"));
    m_client.change_server_url(QUrl(urlString));
}

void MainWindow::resetChats()
{
    if (!m_indexes.empty())
        m_indexes.clear();
    m_indexes["SERVER"] = ui->messagesWidget->addWidget(ui->messages);
    ui->messagesWidget->setCurrentIndex(m_indexes.at("SERVER"));
    chats.clear();
}

void MainWindow::changeLoginState(bool state)
{
    for (auto &action : m_actions)
        action->setEnabled(!state);
    m_actions.last()->setEnabled(state); // Logout action

    ui->addContactButton->setEnabled(state);
    ui->removeContactButton->setEnabled(state);
    ui->sendButton->setEnabled(state);
    ui->lineEdit->setEnabled(state);
}

void MainWindow::displayInfo(QString info)
{
    ui->messages->append(info);
    ui->messagesWidget->setCurrentIndex(m_indexes.at("SERVER"));
    ui->receiver->clear();
}

void MainWindow::on_contacts_itemClicked(QListWidgetItem *item)
{
    ui->receiver->setText(item->text());
    try {
        ui->messagesWidget->setCurrentIndex(m_indexes.at(item->text()));
    }  catch (const std::out_of_range&) {
        displayInfo("ERROR: Unknown contact");
        return;
    }
    QFont normalFont = item->font();
    normalFont.setBold(false);
    item->setFont(normalFont);
}
