#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include <QMainWindow>
#include <QFile>
#include <QListWidget>
#include <QListWidgetItem>
#include <fstream>
#include "../client/Client.h"
#include "login.h"
#include "register.h"
#include <QInputDialog>
#include <QMessageBox>
#include <QTextEdit>
#include <QLocale>

QT_BEGIN_NAMESPACE
namespace Ui { class MainWindow; }
QT_END_NAMESPACE


/**
 * @brief The MainWindow class contains the main chat client GUI
 *
 * It owns its own Client object to send messages through it and keeps
 * username, which changes when login action is successful. ui variable
 * interacts with the graphical components of the program.
 */
class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    MainWindow(QWidget *parent = nullptr);
    ~MainWindow();
    MainWindow &operator=(const MainWindow&) = delete;
    MainWindow(const MainWindow&) = delete;

private slots:
    /**
     * @brief on_clientDisconnected informs user about being disconnected
     */
    void on_clientDisconnected();

    /**
     * @brief on_clientConnected informs user about begin connected to server
     */
    void on_clientConnected();

    /**
     * @brief on_sendButton_clicked sends the message and wipes the input line
     */
    void on_sendButton_clicked();

    /**
     * @brief on_messageReceived after receiving signal from Client, chooses
     * what to do with the message
     * @param message that came from client
     */
    void on_messageReceived(const Message &message);

    /**
     * @brief on_lineEdit_returnPressed enables enter-pressed sending of messages
     */
    void on_lineEdit_returnPressed();

    /**
     * @brief on_actionLogin_triggered opens the Login window
     */
    void on_actionLogin_triggered();

    /**
     * @brief on_loginFinished sends login request
     */
    void on_loginFinished(const QString &name, const QString &password);

    /**
     * @brief on_addContactButton_clicked opens dialog requesting name,
     * if not empty sends request
     */
    void on_addContactButton_clicked();

    /**
     * @brief on_removeContactButton_clicked opens confirmation dialog,
     * if confirmed, sends request of contact removal
     */
    void on_removeContactButton_clicked();

    /**
     * @brief on_actionRegister_triggered opens the Register window
     */
    void on_actionRegister_triggered();

    /**
     * @brief on_registerFinished sends registration request
     */
    void on_registerFinished(const QString &name, const QString &password);

    /**
     * @brief on_actionLogout_triggered sends logout request and
     * changes state to logged out
     */
    void on_actionLogout_triggered();

    /**
     * @brief on_actionSet_server_URL_triggered asks for URL and changes the
     * Client object url
     */
    void on_actionSet_server_URL_triggered();

    /**
     * @brief resetChats clears the messagesWidget of all windows and sets
     * the SERVER one as current
     */
    void resetChats();

    /**
     * @brief loginChange switches between logged in and logged out states
     * by blocking buttons
     */
    void changeLoginState(bool state);

    /**
     * @brief displayInfo displays a message and makes SERVER the current chat window
     * @param info text string
     */
    void displayInfo(QString info);

    /**
     * @brief findContact is helper function to return a contact from contact list
     * @param name to be found
     */
    QListWidgetItem *findContact(QString name);

    /**
     * @brief on_contacts_itemClicked changes the recipient and shows his chat window
     * @param item is the contact which was clicked
     */
    void on_contacts_itemClicked(QListWidgetItem *item);

private:
    std::unique_ptr<Ui::MainWindow> ui;
    Client m_client;
    QString m_username;
    std::map<QString, int> m_indexes;
    QList<QAction *> m_actions;
    std::vector<std::unique_ptr<QTextEdit>> chats;
};
#endif // MAINWINDOW_H
