#include <QtCore/QObject>
#include <QtWebSockets/QWebSocket>
#include <QString>
#include <iostream>
#include "SocketHandler.h"

SocketHandler::SocketHandler(const QUrl &url, QObject *parent) : QObject(parent), m_url(QUrl(url)),
                                                                 m_reconnect_timer(this)
{
    m_web_socket.setParent(this); // this way the socket moves to a different thread together with his parent
    connect(&m_web_socket, &QWebSocket::connected, this, &SocketHandler::on_socket_connected);
    connect(&m_web_socket, &QWebSocket::disconnected, this, &SocketHandler::on_socket_disconnected);
    connect(&m_web_socket, &QWebSocket::textMessageReceived,this, &SocketHandler::on_message_received);
    connect(&m_reconnect_timer, &QTimer::timeout, this, &SocketHandler::try_reconnect);

    m_web_socket.open(QUrl(url));
}

void SocketHandler::on_socket_connected()
{
    m_reconnect_timer.stop();
    m_connected = true;
    if (m_login_msg != "")
    {
        m_web_socket.sendTextMessage(m_login_msg);
    }
    send_messages();
    emit socket_connected();
}


void SocketHandler::on_socket_disconnected()
{
    m_connected = false;
    m_reconnect_timer.start(5000); // 5s
    emit socket_disconnected();
}


void SocketHandler::on_message_added(QString str)
{
    if (m_connected)
    {
        m_web_socket.sendTextMessage(str);
        emit message_sent();
    }
    else
    {
        m_to_send.push_back(str);
    }
}


void SocketHandler::send_messages()
{
    while (!m_to_send.empty() && m_connected)
    {
        m_web_socket.sendTextMessage(m_to_send.front());
        emit message_sent();
        m_to_send.pop_front();
    }
}


void SocketHandler::on_change_server_url(const QUrl &url)
{
    m_url = url;
    m_web_socket.close();
    m_web_socket.open(m_url);
}


void SocketHandler::on_destroy()
{
    disconnect(&m_reconnect_timer, &QTimer::timeout, this, &SocketHandler::try_reconnect);
    m_web_socket.close();
    m_reconnect_timer.stop();
    m_login_msg = "";
}

void SocketHandler::on_recreate()
{
    connect(&m_reconnect_timer, &QTimer::timeout, this, &SocketHandler::try_reconnect);
    try_reconnect();
}

void SocketHandler::on_message_received(QString message)
{
    auto received_msg = Message::from_json_string(message);
    if (received_msg.get_type() == Message::Type::INFO && received_msg.get("success").toBool())
    {
        auto temp = received_msg.get_initial_message().get_raw_string();
        if (Message::from_json_string(temp).get_type() == Message::Type::LOGIN)
        {
            m_login_msg = temp;
        }
    }
    emit message_received(message);
}

void SocketHandler::try_reconnect()
{
    m_web_socket.open(QUrl(m_url));
}







