#include "databasewrapper.h"

#include <iostream>
#include <random>
#include <iostream>
#include <iomanip>

#include <QSqlQuery>
#include <QVariant>
#include <QFile>
#include <QByteArray>
#include <QSqlRecord>
#include <QCryptographicHash>
#include <QUuid>
#include <QCoreApplication>

/*
#include <argon2.h>
*/

bool DatabaseWrapper::containsUser(QString user)
{
    QSqlQuery query(getThreadDatabase());
    query.prepare("SELECT * FROM users WHERE user=:user");
    query.bindValue(":user", user);
    query.exec();
    return query.next();
}

bool DatabaseWrapper::registerUser(QString username, QString password)
{
    constexpr auto SALTLEN{16U};

    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<> distrib(0, 0xFF);
    QByteArray salt;

    for (auto i{0U}; i < SALTLEN; ++i) {
        salt += distrib(gen);
    }

    auto newhash = QCryptographicHash::hash(salt + password.toUtf8(), QCryptographicHash::Algorithm::Sha3_512);

    QSqlQuery query(getThreadDatabase());
    if (!query.prepare("INSERT INTO users VALUES (:user, :hash, :salt)")) {
        qCritical("Failed to prepare query: reason: %s", qPrintable(query.lastError().text()));
    }
    query.bindValue(":user", username);
    query.bindValue(":hash", newhash.toBase64());
    query.bindValue(":salt", salt.toBase64());
    auto res = query.exec();
    if (!res) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }
    return res;
}

bool DatabaseWrapper::authenticateUser(QString username, QString password, QWebSocket *user)
{
    QSqlQuery query(getThreadDatabase());
    if (!query.prepare("SELECT * FROM users WHERE user=:user")) {
        qCritical("Failed to prepare query: reason: %s", qPrintable(query.lastError().text()));
    }
    query.bindValue(":user", username);
    if (!query.exec()) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
        return false;
    }

    if (query.next()) {

        auto hash{QByteArray::fromBase64(query.value("hash").toByteArray())};
        auto salt{QByteArray::fromBase64(query.value("salt").toByteArray())};

        QByteArray cmpHash = QCryptographicHash::hash(salt + password.toUtf8(), QCryptographicHash::Algorithm::Sha3_512);
        if (hash == cmpHash) {
            std::unique_lock lock{m_bimapMutex};
            m_auth_users.emplace(user, username.toStdString());
            return true;
        }
    }
    return false;
}

void DatabaseWrapper::logoutUser(QWebSocket *user)
{
    std::unique_lock lock{m_bimapMutex};
    m_auth_users.remove(user);
}

QString DatabaseWrapper::getUser(QWebSocket *user)
{
    std::unique_lock lock{m_bimapMutex};
    if (m_auth_users.contains(user)) {
        return QString::fromStdString(m_auth_users.getSecond(user));
    }
    return {};
}

QWebSocket *DatabaseWrapper::getSocket(const QString &user)
{
    std::unique_lock lock{m_bimapMutex};
    return m_auth_users.getFirst(user.toStdString());
}

bool DatabaseWrapper::addFriendship(const QString &lhs, const QString &rhs)
{
    if (lhs == rhs) {
        return false;
    }

    QSqlQuery query(getThreadDatabase());
    if (!query.prepare("INSERT INTO friendships VALUES (:lhs, :rhs)")) {
        qCritical("Failed to prepare query: reason: %s", qPrintable(query.lastError().text()));
    }
    query.bindValue(":lhs", lhs);
    query.bindValue(":rhs", rhs);
    auto res = query.exec();
    if (!res) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }
    return res;
}

bool DatabaseWrapper::removeFriendship(const QString &lhs, const QString &rhs)
{
    if (lhs == rhs) {
        return false;
    }

    QSqlQuery query(getThreadDatabase());
    if (!query.prepare("DELETE FROM friendships WHERE (lhs = :lhs AND rhs = :rhs) OR (lhs = :rhs AND rhs = :lhs)")) {
        qCritical("Failed to prepare query: reason: %s", qPrintable(query.lastError().text()));
    }
    query.bindValue(":lhs", lhs);
    query.bindValue(":rhs", rhs);
    auto res = query.exec();
    if (!res) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }
    return res;
}

std::vector<QString> DatabaseWrapper::getFriends(const QString &user)
{
    QSqlQuery query(getThreadDatabase());
    query.prepare(
        "SELECT lhs FROM friendships WHERE rhs = :user UNION "
        "SELECT rhs FROM friendships WHERE lhs = :user");
    query.bindValue(":user", user);
    if (!query.exec()) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }
    std::vector<QString> result;
    while (query.next()) {
        result.push_back(query.value(0).toString());
    }
    return result;
}

void DatabaseWrapper::saveUndeliveredMessage(const QString &user, const QString &msg)
{
    QSqlQuery query(getThreadDatabase());
    query.prepare("INSERT INTO undelivered_messages VALUES (:user, :msg)");
    query.bindValue(":user", user);
    query.bindValue(":msg", msg.toUtf8().toBase64());
    if (!query.exec()) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }
}

std::vector<QString> DatabaseWrapper::getUndeliveredMessages(const QString &user)
{
    std::vector<QString> msgs;
    getThreadDatabase().transaction();
    QSqlQuery query(getThreadDatabase());
    query.prepare("SELECT message FROM undelivered_messages WHERE user = :user");
    query.bindValue(":user", user);
    if (!query.exec()) {
        qCritical("Failed to execute query: reason: %s", qPrintable(query.lastError().text()));
    }

    std::vector<QString> result;
    while (query.next()) {
        result.push_back(QByteArray::fromBase64(query.value(0).toByteArray()));
    }

    QSqlQuery query2(getThreadDatabase());
    if (!query2.prepare("DELETE FROM undelivered_messages WHERE user = :user")) {
        qCritical("Failed to prepare query: reason: %s", qPrintable(query2.lastError().text()));
    }
    query2.bindValue(":user", user);
    query2.exec();
    getThreadDatabase().commit();
    return result;
}

QSqlDatabase DatabaseWrapper::getThreadDatabase()
{
    QString name;
    if (m_threadStorage.hasLocalData()) {
        name = m_threadStorage.localData();
    } else {
        name = "db_" + QUuid::createUuid().toString();
        qDebug("Creating new database connection: %s", qPrintable(name));
        m_threadStorage.setLocalData(name);
    }

    if (QSqlDatabase::contains(name)) {
        return QSqlDatabase::database(name);
    }

    auto db = QSqlDatabase::addDatabase("QSQLITE", name);
    db.setDatabaseName(DatabaseWrapper::filename);
    if (!db.open()) {
        qCritical("Database failed to open: %s", qPrintable(db.lastError().text()));
    }
    return db;
}

QString DatabaseWrapper::filename{};

DatabaseWrapper::DatabaseWrapper() :
    m_db{QSqlDatabase::addDatabase("QSQLITE")}
{
    auto exists = QFile::exists(filename);
    m_db.setDatabaseName(filename);
    if (m_db.open()) {
        if (!exists) {
            qDebug("Creating tables for the database");
            // Create table for user authentication.
            auto query = m_db.exec(
                "CREATE TABLE users ("
                "user TEXT PRIMARY KEY,"
                "hash TEXT NOT NULL,"
                "salt TEXT NOT NULL)"
            );
            if (query.lastError().isValid()) {
                qCritical("Failed to create table users: %s", qPrintable(query.lastError().text()));
            }
            // Create table for friendships.
            auto query2 = m_db.exec(
                "CREATE TABLE friendships ("
                "lhs TEXT,"
                "rhs TEXT,"
                "FOREIGN KEY(lhs) REFERENCES users(userid),"
                "FOREIGN KEY(rhs) REFERENCES users(userid))"
            );
            if (query2.lastError().isValid()) {
                qCritical("Failed to create table users: %s", qPrintable(query2.lastError().text()));
            }
            auto query3 = m_db.exec(
                "CREATE TABLE undelivered_messages ("
                "user TEXT,"
                "message TEXT NOT NULL,"
                "FOREIGN KEY(user) REFERENCES users(userid))"
            );
            if (query3.lastError().isValid()) {
                qCritical("Failed to create table users: %s", qPrintable(query3.lastError().text()));
            }

        }
    } else {
        qCritical("Database failed to open: %s", qPrintable(m_db.lastError().text()));
    }

}