#include "worker.h"

#include "databasewrapper.h"
#include "chatserver.h"

#include <iostream>
#include <QWebSocket>
#include <QMetaObject>

Worker::Worker(ChatServer &server) : Worker(1, server) {}

Worker::Worker(unsigned numberOfThreads, ChatServer &server) :
    m_numberOfThreads{numberOfThreads},
    m_server{server}
{
    qRegisterMetaType<QWebSocket*>("QWebSocket*");
    qRegisterMetaType<std::promise<qint64>*>("std::promise<qint64>*");

    for (auto i{0U}; i < m_numberOfThreads; ++i) {
        m_threads.emplace_back([this] { work(); });
    }
}

Worker::~Worker()
{
    m_terminate = true;
    m_condition.notify_all();
    for (auto &v : m_threads) {
        v.join();
    }
}

void Worker::addJob(Job job)
{
    {
        std::unique_lock lock(m_mutex);
        m_jobs.push(std::move(job));
    }
    m_condition.notify_one();
}

void Worker::work()
{

    while (true) {
        std::unique_lock lock(m_mutex);
        m_condition.wait(lock, [this] { return !m_jobs.empty() || m_terminate; });
        if (m_terminate) {
            return;
        }
        Job job = m_jobs.front();
        m_jobs.pop();
        lock.unlock();
        m_condition.notify_one();
        processJob(std::move(job));
    }
}

void Worker::processJob(Job job) const
{
    using namespace std::string_literals;

    const auto &msg = job.msg;
    std::vector<std::pair<Message, QWebSocket*>> responses;
    switch (msg.get_type())
    {
    case Message::Type::LOGIN : {
        auto ok = DatabaseWrapper::instance().authenticateUser(msg.get("sender").toString(), msg.get("password").toString(), job.socket);
        auto response = QString::fromStdString(ok ? "Login successful"s : "Login unsuccessful"s);
        auto resp = Message::create_info_message(msg.get_recipient(), msg, ok, response);
        responses.emplace_back(resp, job.socket);

        if (ok) {
            auto friends = DatabaseWrapper::instance().getFriends(msg.get_sender());
            QJsonArray array;
            std::copy(friends.begin(), friends.end(), std::back_inserter(array));
            responses.emplace_back(Message::create_contact_list_response(array), job.socket);

            auto msgs = DatabaseWrapper::instance().getUndeliveredMessages(msg.get_sender());
            for (const auto &msg : msgs) {
                responses.emplace_back(Message::from_json_string(msg), job.socket);
            }
        }
        break;
    }
    case Message::Type::LOGOUT : {
        DatabaseWrapper::instance().logoutUser(job.socket);
        auto resp = Message::create_info_message(msg.get_recipient(), msg, true, "Logout successful");
        responses.emplace_back(resp, job.socket);
        break;
    }
    case Message::Type::REGISTER : {
        auto ok = DatabaseWrapper::instance().registerUser(msg.get("sender").toString(), msg.get("password").toString());
        auto response = QString::fromStdString(ok ? "Registration successful"s : "Registration unsuccessful"s);
        auto resp = Message::create_info_message(msg.get_recipient(), msg, ok, response);
        responses.emplace_back(resp, job.socket);
        break;
    }
    case Message::Type::TEXT : {
        auto user = DatabaseWrapper::instance().getUser(job.socket);
        if (user.isEmpty() || user != msg.get_sender()) {
            qWarning("User tried to send message, but login info doesn't match. expected: %s, actual: %s, socket: %s", qPrintable(msg.get_sender()), qPrintable(user), qPrintable(ChatServer::getIdentifier(job.socket)));
            break;
        }
        auto resp = msg;
        auto *socket = DatabaseWrapper::instance().getSocket(msg.get_recipient());
        if (socket != nullptr) {
            responses.emplace_back(resp, socket);
        } else {
            DatabaseWrapper::instance().saveUndeliveredMessage(msg.get_recipient(), msg.get_raw_string());
        }
        break;
    }
    case Message::Type::CONTACT_LIST_REQUEST : {
        auto user = DatabaseWrapper::instance().getUser(job.socket);
        if (user.isEmpty() || user != msg.get_sender()) {
            qWarning("User tried to list friends, but login info doesn't match. expected: %s, actual: %s, socket: %s", qPrintable(msg.get_sender()), qPrintable(user), qPrintable(ChatServer::getIdentifier(job.socket)));
            break;
        }
        auto friends = DatabaseWrapper::instance().getFriends(msg.get_sender());
        QJsonArray array;
        std::copy(friends.begin(), friends.end(), std::back_inserter(array));
        auto resp = Message::create_contact_list_response(array);
        responses.emplace_back(resp, job.socket);
        break;
    }
    case Message::Type::FRIEND_REQUEST : {
        auto user = DatabaseWrapper::instance().getUser(job.socket);
        if (user.isEmpty() || user != msg.get_sender()) {
            qWarning("User tried to add friends, but login info doesn't match. expected: %s, actual: %s, socket: %s", qPrintable(msg.get_sender()), qPrintable(user), qPrintable(ChatServer::getIdentifier(job.socket)));
            break;
        }
        DatabaseWrapper::instance().addFriendship(msg.get_sender(), msg.get_recipient());

        auto friends1 = DatabaseWrapper::instance().getFriends(msg.get_sender());
        auto friends2 = DatabaseWrapper::instance().getFriends(msg.get_recipient());
        QJsonArray array1;
        std::copy(friends1.begin(), friends1.end(), std::back_inserter(array1));
        QJsonArray array2;
        std::copy(friends2.begin(), friends2.end(), std::back_inserter(array2));
        auto resp1 = Message::create_contact_list_response(array1);
        auto resp2 = Message::create_contact_list_response(array2);

        responses.emplace_back(resp1, DatabaseWrapper::instance().getSocket(msg.get_sender()));
        responses.emplace_back(resp2, DatabaseWrapper::instance().getSocket(msg.get_recipient()));
        break;
    }
    case Message::Type::UNFRIEND_REQUEST : {
        auto user = DatabaseWrapper::instance().getUser(job.socket);
        if (user.isEmpty() || user != msg.get_sender()) {
            qWarning("User tried to unfriend someone, but login info doesn't match. expected: %s, actual: %s, socket: %s", qPrintable(msg.get_sender()), qPrintable(user), qPrintable(ChatServer::getIdentifier(job.socket)));
            break;
        }
        DatabaseWrapper::instance().removeFriendship(msg.get_sender(), msg.get_recipient());

        auto friends1 = DatabaseWrapper::instance().getFriends(msg.get_sender());
        auto friends2 = DatabaseWrapper::instance().getFriends(msg.get_recipient());
        QJsonArray array1;
        std::copy(friends1.begin(), friends1.end(), std::back_inserter(array1));
        QJsonArray array2;
        std::copy(friends2.begin(), friends2.end(), std::back_inserter(array2));
        auto resp1 = Message::create_contact_list_response(array1);
        auto resp2 = Message::create_contact_list_response(array2);

        responses.emplace_back(resp1, DatabaseWrapper::instance().getSocket(msg.get_sender()));
        responses.emplace_back(resp2, DatabaseWrapper::instance().getSocket(msg.get_recipient()));
        break;
    }
    default:
        break;
    }

    for (auto &v : responses) {
        if (v.second) {
            std::promise<qint64> promise;
            auto bytes = promise.get_future();
            auto msgString = v.first.get_raw_string();
            QMetaObject::invokeMethod(
                &m_server,
                "sendMessage",
                Qt::QueuedConnection,
                Q_ARG(QString, msgString),
                Q_ARG(QWebSocket*, v.second),
                Q_ARG(std::promise<qint64>*, &promise)
            );

            auto bytesWritten = bytes.get();
            if (bytesWritten == msgString.size()) {
                qDebug("Message successfuly sent. size: %lld bytes", bytesWritten);
            } else {
                if (v.first.get_type() == Message::Type::TEXT) {
                    DatabaseWrapper::instance().saveUndeliveredMessage(msg.get_recipient(), msg.get_raw_string());
                }
                qDebug("Message was not successfuly send. expected: %d, actual: %lld", msgString.size(), bytesWritten);
            }
        }
    }
}