#pragma once

#include <QtCore>


/**
 * Each message sent between a client and a server is sent via a websocket as a string.
 * Message has a JSON form, is a JSON object, and has a "type" key representing some functionality
 *
 * Type::TEXT                     represents a simple exchange between two clients,
 *                                fields: "type", "sender", "recipient", "message"
 *
 * Type::LOGIN                    tells server to search for "name" in its registered users list and verify the "password"
 *                                fields: "type", "name", "password"
 *
 * Type::REGISTER                 tells server to create a new user with "name" and "password"
 *                                fields: "type", "name", "password"
 *
 * Type::LOGOUT                   informs server that the client is logging out
 *                                fields: "type", "sender"
 *
 * Type::FRIEND_REQUEST           a friends request
 *                                fields: "type", "sender", "recipient"
 *
 * Type::UNFRIEND_REQUEST         a request to delete a contact from your list
 *                                fields: "type", "sender", "recipient"
 *
 * Type::FRIEND_REQUEST_RESPONSE  a response to FRIEND_REQUEST
 *                                fields: "type", "sender", "recipient", "accepted"
 *
 * Type::CONTACT_LIST_REQUEST     requests a list of contacts from the server
 *                                fields: "type", "sender"
 *
 * Type::CONTACT_LIST_RESPONSE    a response from server containing json array of contacts
 *                                fields: "type", "friends"
 *
 * Type::INFO                     a message sent only from server to client
 *                                informing client whether the a previous action "initial_message" was successful or not
 *                                fields: "type", "recipient", "initial_message", "success", "info"
 */

struct Message
{
    enum class Type : int
    {
        TEXT = 0,
        LOGIN = 1,
        LOGOUT = 2,
        REGISTER = 3,
        FRIEND_REQUEST = 4,
        UNFRIEND_REQUEST = 5,
        FRIEND_REQUEST_RESPONSE = 6,
        CONTACT_LIST_REQUEST = 7,
        CONTACT_LIST_RESPONSE = 8,
        INFO = 9
    };

    /**
     * Creates Message from QByteArray, which must be in json format
     * and must contain required fields as mentioned above
     */
    static Message from_json_byte(const QByteArray& array);
    static Message from_json_string(const QString& str) { return from_json_byte(str.toUtf8()); }

    static Message create_text_message(const QString& sender, const QString& recipient, const QString& msg);
    static Message create_login_message(const QString& sender, const QString& password);
    static Message create_logout_message(const QString &sender);
    static Message create_register_message(const QString& sender, const QString& password);
    static Message create_friend_request(const QString& sender, const QString& friend_);
    static Message create_unfriend_request(const QString& sender, const QString& friend_);
    static Message create_friend_request_response(const QString& sender, const QString& recipient, bool accepted);
    static Message create_info_message(const QString& recipient, const Message& initial_msg, bool success, QString info);
    static Message create_contact_list_request(const QString& sender);
    static Message create_contact_list_response(const QJsonArray& friends);

    Type get_type() const { return Message::Type(m_scheme["type"].toInt()); }
    QJsonValue get(const QString& key) const { return m_scheme[key]; }
    QDateTime get_time() const { return QDateTime::fromString(m_scheme["time"].toString(), dateTimeFormat); }
    QString get_recipient() const { return m_scheme["recipient"].toString(); }
    QString get_friend() const { return m_scheme["recipient"].toString(); }
    QString get_sender() const { return m_scheme["sender"].toString(); }
    QString get_message() const { return m_scheme["message"].toString(); }
    Message get_initial_message() const { return Message::from_json_string(m_scheme["initial_message"].toString()); }
    QString get_raw_string() const { return QString(QJsonDocument(m_scheme).toJson()); }
    QByteArray get_raw_bytes() const { return QJsonDocument(m_scheme).toJson(); }
    QJsonArray get_contact_list() const { return m_scheme["friends"].toArray(); }

    //bool operator==(const Message& other) const = default;
    bool equals_without_time(const Message& other) const;
    Message() = default;


    static const QString dateTimeFormat;

protected:
    explicit Message(const QJsonObject& scheme) : m_scheme(scheme)
    {
        if (!m_scheme.contains("time"))
            m_scheme.insert("time", QDateTime::currentDateTime().toString(dateTimeFormat));
    }

    QJsonObject m_scheme;
};
Q_DECLARE_METATYPE(Message)

